/*
    ipodfs - a user-space filesystem for accessing iPods
    Copyright (C) 2006 Richard Dawe

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    iTunes and iPod are trademarks of Apple. This product is not
    supported/written/published by Apple.
*/

#include <ctype.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/vfs.h>
#include <sys/statvfs.h>
#include <sys/param.h>
#include <getopt.h>
#include <pthread.h>

#define FUSE_USE_VERSION 25
#include <fuse.h>

#include <gpod/itdb.h>

#include "ipodfs_version.h"

#ifndef DEV_BSIZE
#error DEV_BSIZE is not defined! Does your system default to 512 bytes per block?
#endif

#define S_IRWX(x) ((x) & (S_IRWXU | S_IRWXG | S_IRWXO))

typedef struct {
  pthread_mutex_t itunes_mutex;

  Itdb_iTunesDB *itunes;
  const char *ipod_path;
  struct statvfs ipod_statvfs;
} ipodfs_context;

static const char DEFAULT_IPOD_PATH[] = "/media/IPOD";

static ipodfs_context initial_ctx;

static ipodfs_context *
ipodfs_get_context (void)
{
  ipodfs_context *ctx = NULL;
  struct fuse_context *fuse_ctx = fuse_get_context();

  if (fuse_ctx && fuse_ctx->private_data)
    ctx = (ipodfs_context *) fuse_ctx->private_data;

  return ctx;
}

/*!
 * This function assumes that the caller already holds the itunes mutex.
 * It will not lock the DB itself.
 *
 * XXX: Something better than O(N).
*/

static int
resolve (Itdb_iTunesDB *itunes,
	 const char *path,
	 char **resolved_path,
	 Itdb_Track **resolved_track)
{
  *resolved_path = NULL;
  *resolved_track = NULL;

  if (strcmp(path, "/") != 0)
    {
      int n_tracks;
      int i;

      /* Track */
      ++path;

      /* Find track */
      n_tracks = itdb_tracks_number(itunes);
      for (i = 0; i < n_tracks; ++i)
	{
	  Itdb_Track *track = g_list_nth_data(itunes->tracks, i);
	  char *track_path = itdb_filename_on_ipod(track);
	  char *filename = NULL;

	  filename = strrchr(track_path, '/');
	  if (filename)
	    ++filename;
	  else
	    filename = track_path;

	  if (strcmp(filename, path) == 0)
	    {
	      *resolved_path = track_path;
	      *resolved_track = track;
	      break;
	    }

	  free(track_path);
	}
    }

  return *resolved_path && *resolved_track;
}

/*!
 * This function assumes that the caller already holds the itunes mutex.
 * It will not lock the DB itself.
 */

static int
write_dbs (Itdb_iTunesDB *db)
{
  GError *err = NULL;
  int ok = 1;

  itdb_write(db, &err);
  if (err != NULL)
    {
      fprintf (stderr, "Unable to write iTunes database: %s\n", err->message);
      g_error_free (err);
      err = NULL;
      ok = 0;
    }

  itdb_shuffle_write(db, &err);
  if (err != NULL)
    {
      fprintf (stderr, "Unable to write iTunes shuffle database: %s\n", err->message);
      g_error_free (err);
      err = NULL;
      ok = 0;
    }

  return ok;
}

static int
ipodfs_getattr (const char *path, struct stat *sbuf)
{
  ipodfs_context *ctx = ipodfs_get_context();
  int cur_errno = ENOENT;
  int found = 0;

  if (strcmp(path, "/") == 0)
    {
      struct stat root_sbuf;

      memset(&root_sbuf, 0, sizeof(root_sbuf));
      if (stat(ctx->ipod_path, &root_sbuf) == -1)
	{
	  cur_errno = errno;
	}
      else
	{
	  sbuf->st_dev = root_sbuf.st_dev;
	  sbuf->st_ino = 0;
	  sbuf->st_mode = S_IFDIR | S_IRWX(root_sbuf.st_mode);
	  sbuf->st_nlink = 1;
	  sbuf->st_uid = getuid();
	  sbuf->st_gid = getgid();
	  sbuf->st_size = 0;
	  sbuf->st_blksize = ctx->ipod_statvfs.f_bsize;
	  sbuf->st_blocks = 1;
	  sbuf->st_atime = root_sbuf.st_atime;
	  sbuf->st_mtime = root_sbuf.st_mtime;
	  sbuf->st_ctime = root_sbuf.st_ctime;

	  found = 1;
	}
    }
  else if (*path == '/')
    {
      /* Track */
      char *track_path = NULL;
      Itdb_Track *track = NULL;

      /* Grab mutex */
      pthread_mutex_lock(&ctx->itunes_mutex);

      if (resolve(ctx->itunes, path, &track_path, &track))
	{
	  struct stat track_sbuf;
	  int ret;

	  memset(&track_sbuf, 0, sizeof(track_sbuf));
	  ret = stat(track_path, &track_sbuf);
	  if (ret == -1)
	    {
	      cur_errno = errno;
	    }
	  else
	    {
	      sbuf->st_dev = track_sbuf.st_dev;
	      sbuf->st_ino = track->id;
	      sbuf->st_mode = S_IFREG | S_IRWX(track_sbuf.st_mode);
	      sbuf->st_nlink = 1;
	      sbuf->st_uid = getuid();
	      sbuf->st_gid = getgid();
	      sbuf->st_size = track_sbuf.st_size;
	      sbuf->st_blksize = ctx->ipod_statvfs.f_bsize;
	      sbuf->st_blocks = (track_sbuf.st_size + DEV_BSIZE) / DEV_BSIZE;
	      sbuf->st_atime = track_sbuf.st_atime;
	      sbuf->st_mtime = track_sbuf.st_mtime;
	      sbuf->st_ctime = track_sbuf.st_ctime;

	      found = 1;
	    }
	}

      /* Release mutex */
      pthread_mutex_unlock(&ctx->itunes_mutex);

      if (track_path)
	free(track_path);
    }

  return found ? 0 : -cur_errno;
}

static int
ipodfs_mknod (const char *path, mode_t mode, dev_t rdev)
{
  ipodfs_context *ctx = ipodfs_get_context();
  int cur_errno = ENOENT;
  int found = 0;
  int ret = 0;

  if (strcmp(path, "/") == 0)
    {
      /* Root */
      cur_errno = EISDIR;
    }
  else
    {
      /* Track */

      /* XXX: Guess the file type */
      /* XXX: Mark the file as new, so we add it to the DB on release */
      if (!S_ISREG(mode) || (strchr(path + 1, '/') != NULL))
	{
	  /*
	   * Only allow regular files. Only allow files in the current
	   * directory.
	   */
	  cur_errno = EPERM;
	}
      else
	{
	  int n_dirs;
	  char track_path[PATH_MAX];
	  char track_ipod_path[PATH_MAX];
	  Itdb_Track *track = NULL;
	  int n;

	  ++path; /* Skip leading '/' */

	  /* Grab mutex */
	  pthread_mutex_lock(&ctx->itunes_mutex);

	  /* Generate a filename in one of the Fxx directories */
	  n_dirs = itdb_musicdirs_number(ctx->itunes);
	  n = rand() % n_dirs;
	  snprintf(track_path, sizeof(track_path),
		   "%s/iPod_Control/Music/F%02d/%s", ctx->ipod_path, n, path);
	  snprintf(track_ipod_path, sizeof(track_ipod_path),
		   "iPod_Control:Music:F%02d:%s", n, path);

	  /* XXX: Add to the master playlist. */
	  track = itdb_track_new();
	  if (!track)
	    {
	      cur_errno = ENOMEM;
	    }
	  else
	    {
	      char *ext = NULL;

	      /*
	       * Use the file extension as the description. Crap,
	       * but this is the best way of getting the filetype
	       * set up at the start. We can do better once some data
	       * has been read to the file - e.g.: using ID3 tags.
	       */
	      ext = strrchr(path, '.');
	      if (ext)
		{
		  ++ext;
		  ext = strdup(ext);
		}
	      if (ext)
		{
		  char *p;
		  for (p = ext; *p; p++) { *p = toupper(*p); }
		}

	      /* XXX: Other fields? E.g.: size? */
	      track->ipod_path = strdup(track_ipod_path);
	      track->description = ext;

	      ret = mknod(track_path, mode, rdev);
	      if (ret == -1)
		{
		  cur_errno = errno;
		}
	      else
		{
		  Itdb_Playlist *master_pl = itdb_playlist_mpl(ctx->itunes);

		  itdb_track_add(ctx->itunes, track, -1);
		  itdb_playlist_add_track(master_pl, track, -1);

		  /* Write out the DBs */
		  if (write_dbs(ctx->itunes))
		    found = 1;
		}
	    }

	  /* Release mutex */
	  pthread_mutex_unlock(&ctx->itunes_mutex);
	}
    }

  return found ? 0 : -cur_errno;
}

static int
ipodfs_unlink (const char *path)
{
  ipodfs_context *ctx = ipodfs_get_context();
  int cur_errno = ENOENT;
  int found = 0;

  if (strcmp(path, "/") == 0)
    {
      /* Root */
      cur_errno = EISDIR;
    }
  else
    {
      /* Track */
      char *track_path = NULL;
      Itdb_Track *track = NULL;

      /* Grab mutex */
      pthread_mutex_lock(&ctx->itunes_mutex);

      if (resolve(ctx->itunes, path, &track_path, &track))
	{
	  int n_playlists;
	  int i;

	  /* Remove the track from any playlists */
	  n_playlists = itdb_playlists_number(ctx->itunes);
	  for (i = 0; i < n_playlists; i++)
	    {
	      Itdb_Playlist *pl = g_list_nth_data(ctx->itunes->playlists, i);

	      itdb_playlist_remove_track(pl, track);
	    }

	  /* Remove the track from the DB */
	  itdb_track_remove(track);

	  /* Write out the DBs */
	  if (write_dbs(ctx->itunes))
	    {
	      /* Finally physically remove the file */
	      int ret = unlink(track_path);
	      if (ret == -1)
		cur_errno = errno;
	      else
		found = 1;
	    }
	}

      /* Release mutex */
      pthread_mutex_unlock(&ctx->itunes_mutex);

      if (track_path)
	free(track_path);
    }

  return found ? 0 : -cur_errno;
}

static int
replace_last_component (const char *str, const char sep, const char *replacement, char *buf, const size_t buflen)
{
  char *p = NULL;

  strncpy(buf, str, buflen);
  buf[buflen - 1] = '\0';

  p = strrchr(buf, sep);
  if (p)
    {
      /*
       * Path separator - replace last path component
       * with new filename.
       */
      size_t spc;

      ++p;
      spc = buflen - (ptrdiff_t) (p - buf);
      strncpy(p, replacement, spc);
      p[spc - 1] = '\0';
    }
  else
    {
      /* No path separator -- weird. Just use new filename. */
      strncpy(buf, replacement, buflen);
    }

  /* XXX: Detect truncation */
  return 0;
}

static int
ipodfs_rename (const char *oldpath, const char *newpath)
{
  ipodfs_context *ctx = ipodfs_get_context();
  int cur_errno = ENOENT;
  int found = 0;

  if (strcmp(oldpath, "/") == 0)
    {
      /* Root */
      cur_errno = EPERM;
    }
  else
    {
      /* Track */
      char *track_oldpath = NULL;
      Itdb_Track *track = NULL;

      /* Grab mutex */
      pthread_mutex_lock(&ctx->itunes_mutex);

      if (resolve(ctx->itunes, oldpath, &track_oldpath, &track))
	{
	  char track_newpath[PATH_MAX];
	  char track_ipod_newpath[PATH_MAX];

	  /* Skip over root directory. */
	  ++newpath;

	  /* Build new names. */
	  replace_last_component(track_oldpath, '/', newpath,
				 track_newpath, sizeof(track_newpath));

	  replace_last_component(track->ipod_path, ':', newpath,
				 track_ipod_newpath, sizeof(track_ipod_newpath));

	  /* Rename the file */
	  if (rename(track_oldpath, track_newpath) == 0)
	    {
	      /* Update the playlist entry. */
	      free(track->ipod_path);
	      track->ipod_path = strdup(track_ipod_newpath);
	      if (!track->ipod_path)
		cur_errno = ENOMEM;

	      /* Write out the DBs */
	      if (track->ipod_path && write_dbs(ctx->itunes))
		found = 1;
	    }
	  else
	    {
	      cur_errno = errno;
	    }
	}

      /* Release mutex */
      pthread_mutex_unlock(&ctx->itunes_mutex);

      if (track_oldpath)
	free(track_oldpath);
    }

  return found ? 0 : -cur_errno;
}

static int
ipodfs_chmod (const char *path, mode_t mode)
{
  ipodfs_context *ctx = ipodfs_get_context();
  int cur_errno = ENOENT;
  int found = 0;

  if (strcmp(path, "/") == 0)
    {
      /* Root */
      cur_errno = EPERM;
    }
  else
    {
      /* Track */
      char *track_path = NULL;
      Itdb_Track *track = NULL;

      /* Grab mutex */
      pthread_mutex_lock(&ctx->itunes_mutex);

      if (resolve(ctx->itunes, path, &track_path, &track))
	{
	  int ret = chmod(track_path, mode);
	  if (ret == -1)
	    cur_errno = errno;
	  else
	    found = 1;
	}

      /* Release mutex */
      pthread_mutex_unlock(&ctx->itunes_mutex);

      if (track_path)
	free(track_path);
    }

  return found ? 0 : -cur_errno;
}

static int
ipodfs_chown (const char *path, uid_t uid, gid_t gid)
{
  int cur_errno = 0;

  /* XXX: Can the filesystem be mounted for write for other users? */
  if (uid != getuid() || gid != getgid())
    cur_errno = EPERM;

  return -cur_errno;
}

static int
ipodfs_truncate (const char *path, off_t size)
{
  ipodfs_context *ctx = ipodfs_get_context();
  int cur_errno = ENOENT;
  int found = 0;

  if (strcmp(path, "/") == 0)
    {
      /* Root */
      cur_errno = EISDIR;
    }
  else
    {
      /* Track */
      char *track_path = NULL;
      Itdb_Track *track = NULL;

      /* Grab mutex */
      pthread_mutex_lock(&ctx->itunes_mutex);

      if (resolve(ctx->itunes, path, &track_path, &track))
	{
	  int ret = truncate(track_path, size);
	  if (ret == -1)
	    cur_errno = errno;
	  else
	    found = 1;
	}

      /* Release mutex */
      pthread_mutex_unlock(&ctx->itunes_mutex);

      if (track_path)
	free(track_path);
    }

  return found ? 0 : -cur_errno;
}

static int
ipodfs_utime (const char *path, struct utimbuf *buf)
{
  ipodfs_context *ctx = ipodfs_get_context();
  int cur_errno = ENOENT;
  int found = 0;
  char *track_path = NULL;
  Itdb_Track *track = NULL;

  /* Grab mutex */
  pthread_mutex_lock(&ctx->itunes_mutex);

  if (   (strcmp(path, "/") == 0)
      || resolve(ctx->itunes, path, &track_path, &track))
    {
      if (utime(track_path, buf) == -1)
	cur_errno = errno;
      else
	found = 1;
    }

  /* Release mutex */
  pthread_mutex_unlock(&ctx->itunes_mutex);

  return found ? 0 : -cur_errno;
}

static int
ipodfs_open (const char *path, struct fuse_file_info *fi)
{
  ipodfs_context *ctx = ipodfs_get_context();
  int cur_errno = ENOENT;
  int found = 0;

  if (strcmp(path, "/") == 0)
    {
      /* Root */
      cur_errno = EISDIR;
    }
  else
    {
      /* Track */
      char *track_path = NULL;
      Itdb_Track *track = NULL;

      /* Grab mutex */
      pthread_mutex_lock(&ctx->itunes_mutex);

      if (resolve(ctx->itunes, path, &track_path, &track))
	{
	  /* Existing track */
	  /* XXX: What should we do if the file is already open? */
	  int fd = open(track_path, fi->flags);
	  if (fd >= 0)
	    {
	      fi->fh = fd; /* XXX: fd == 0 => need to dup to get non-zero? */

	      found = 1;
	    }
	  else
	    {
	      fi->fh = 0;

	      cur_errno = errno;
	    }
	}

      /* Release mutex */
      pthread_mutex_unlock(&ctx->itunes_mutex);

      if (track_path)
	free(track_path);
    }

  return found ? 0 : -cur_errno;
}

static int
ipodfs_read (const char *path, char *buf, size_t size, off_t offset,
	     struct fuse_file_info *fi)
{
  int cur_errno = EBADF;
  int found = 0;
  int ret = 0;

  if (fi->fh)
    {
      ret = lseek(fi->fh, offset, SEEK_SET);
      if (ret == -1)
	{
	  cur_errno = errno;
	}
      else
	{
	  ret = read(fi->fh, buf, size);
	  if (ret == -1)
	    cur_errno = errno;
	  else
	    found = 1;
	}
    }

  return found ? ret : -cur_errno;
}

static int
ipodfs_write (const char *path, const char *buf, size_t size,
	      off_t offset, struct fuse_file_info *fi)
{
  int cur_errno = EBADF;
  int found = 0;
  int ret = 0;

  if (fi->fh)
    {
      ret = lseek(fi->fh, offset, SEEK_SET);
      if (ret == -1)
	{
	  cur_errno = errno;
	}
      else
	{
	  ret = write(fi->fh, buf, size);
	  if (ret == -1)
	    cur_errno = errno;
	  else
	    found = 1;
	}
    }

  return found ? ret : -cur_errno;
}

static int
ipodfs_statfs (const char *path, struct statvfs *sbuf)
{
  ipodfs_context *ctx = ipodfs_get_context();
  int ret;

  ret = statvfs(ctx->ipod_path, sbuf);
  if (ret == -1)
    ret = -errno;

  return ret;
}

static int
ipodfs_release (const char *path, struct fuse_file_info *fi)
{
  int cur_errno = ENOENT;
  int found = 0;

  if (fi->fh)
    {
      close(fi->fh);
      fi->fh = 0;

      found = 1;
    }
  /* XXX: Something needed for other cases? */

  return found ? 0 : -cur_errno;
}

static int
ipodfs_fsync (const char *path, int isdatasync, struct fuse_file_info *fi)
{
  int cur_errno = EBADF;
  int found = 0;

  if (fi->fh)
    {
      int ret = 0;

      if (isdatasync)
	ret = fdatasync(fi->fh);
      else
	ret = fsync(fi->fh);

      if (ret == -1)
	cur_errno = errno;
      else
	found = 1;
    }
  /* XXX: Something needed for other cases? */

  return found ? 0 : -cur_errno;  
}

static int
ipodfs_readdir (const char *path,
		void *buf,
		fuse_fill_dir_t filler,
		off_t offset,
		struct fuse_file_info *fi)
{
  ipodfs_context *ctx = ipodfs_get_context();
  int found = 0;
  int i;

  /* Grab mutex */
  pthread_mutex_lock(&ctx->itunes_mutex);

  if (strcmp(path, "/") == 0)
    {
      const int n_tracks = itdb_tracks_number(ctx->itunes);

      for (i = 0; i < n_tracks; ++i)
	{
	  Itdb_Track *track = g_list_nth_data(ctx->itunes->tracks, i);
	  char *track_path = itdb_filename_on_ipod(track);
	  char *filename = NULL;
	  struct stat sbuf;

	  filename = strrchr(track_path, '/');
	  if (filename)
	    ++filename;
	  else
	    filename = track_path;
	  
	  memset(&sbuf, 0, sizeof(sbuf));
	  sbuf.st_ino = track->id;
	  sbuf.st_mode = S_IFREG;
	  if (filler(buf, filename, &sbuf, 0))
	    {
	      free(track_path);
	      break;
	    }

	  free(track_path);
	}

      found = 1;
    }

  /* Release mutex */
  pthread_mutex_unlock(&ctx->itunes_mutex);

  return found ? 0 : -ENOENT;
}

static void *
ipodfs_init (void)
{
  ipodfs_context *ctx;

  ctx = malloc(sizeof(*ctx));
  if (ctx)
    {
      memcpy(ctx, &initial_ctx, sizeof(*ctx));

      pthread_mutex_init(&ctx->itunes_mutex, NULL);
    }

  return ctx;
}

static void
ipodfs_destroy (void *ptr)
{
  ipodfs_context *ctx = (ipodfs_context *) ptr;

  if (ctx)
    pthread_mutex_destroy(&ctx->itunes_mutex);

  if (ptr)
    free(ptr);
}

static struct fuse_operations ipodfs_operations = {
  ipodfs_getattr,
  /* ipodfs_readlink */ NULL,
  /* ipodfs_getdir */ NULL,
  ipodfs_mknod,
  /* ipodfs_mkdir */ NULL,
  ipodfs_unlink,
  /* ipodfs_rmdir */ NULL,
  /* ipodfs_symlink */ NULL,
  ipodfs_rename,
  /* ipodfs_link */ NULL,
  ipodfs_chmod,
  ipodfs_chown,
  ipodfs_truncate,
  ipodfs_utime,
  ipodfs_open,
  ipodfs_read,
  ipodfs_write,
  ipodfs_statfs,
  /* ipodfs_flush */ NULL,
  ipodfs_release,
  ipodfs_fsync,
  /* ipodfs_setxattr */ NULL,
  /* ipodfs_getxattr */ NULL,
  /* ipodfs_listxattr */ NULL,
  /* ipodfs_removexattr */ NULL,
  /* ipodfs_opendir */ NULL,
  ipodfs_readdir,
  /* ipodfs_releasedir */ NULL,
  /* ipodfs_fsyncdir */ NULL,
  ipodfs_init,
  ipodfs_destroy,
  /* ipodfs_access */ NULL,
  /* ipodfs_create */ NULL,
  /* ipodfs_ftruncate */ NULL,
  /* ipodfs_fgetattr */ NULL,
};

static void
version (void)
{
  fprintf(stderr,
	  "%s - a user-space filesystem for accessing iPods\n"
	  "Copyright (C) %s\n",
	  PACKAGE_STRING,
	  PACKAGE_COPYRIGHT);
}

static void
help (const char *program_name)
{
  version();
  fprintf(stderr,
"%s comes with ABSOLUTELY NO WARRANTY. This is free software,\n"
"and you are welcome to redistribute it under certain conditions;\n"
"for details run ``%s --license''.\n"
"\n"
"iTunes and iPod are trademarks of Apple. This product is not\n"
"supported/written/published by Apple.\n"
"\n"
"Usage: %s [OPTIONS] -- [FUSE-OPTIONS] MOUNT-POINT\n"
"\n"
"Options:\n"
"  -i IPOD-PATH, --ipod=IPOD-PATH       where the iPod is mounted\n"
"                                       (defaults to ``%s'')\n"
"  -h, --help                           display this help\n"
"  -l, --license                        display the license\n"
"  -v, --version                        display the version\n"
"\n"
"FUSE options:\n"
"  -f                                   foreground; don't run as daemon\n"
"  -d                                   foreground in debug mode\n"
"  -s                                   single-threaded\n"
"\n"
"Report bugs to <%s>.\n",
	  PACKAGE_NAME,
	  program_name,
	  program_name,
	  DEFAULT_IPOD_PATH,
	  PACKAGE_BUGREPORT);
}

static void
license (void)
{
  version();
  fprintf(stderr,
"\n"
"This program is free software; you can redistribute it and/or modify\n"
"it under the terms of the GNU General Public License as published by\n"
"the Free Software Foundation; either version 2 of the License, or\n"
"(at your option) any later version.\n"
"\n"
"This program is distributed in the hope that it will be useful,\n"
"but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
"GNU General Public License for more details.\n"
"\n"
"You should have received a copy of the GNU General Public License\n"
"along with this program; if not, write to the Free Software\n"
"Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA\n"
"\n"
"iTunes and iPod are trademarks of Apple. This product is not\n"
"supported/written/published by Apple.\n");
}

int
main (int argc, char *argv[])
{
  const char *program_name = argv[0];
  int show_help = 0;
  int show_license = 0;
  int show_version = 0;
  Itdb_iTunesDB *itunes = NULL;
  const char *ipod_path = NULL;
  struct statvfs ipod_statvfs;
  const char *p = NULL;
  GError *err = NULL;
  int ok = 1;
  int fuse_argc = 0;
  char **fuse_argv = NULL;

  p = strrchr(program_name, '/');
  if (p)
    {
      ++p;
      program_name = p;
    }

  /* Parse command-line arguments */
  while (1)
    {
      static struct option long_options[] = {
	{ "help", 0, NULL, 'h' },
	{ "ipod", 1, NULL, 'i' },
	{ "license", 0, NULL, 'l' },
	{ "version", 0, NULL, 'v' },

	/* Terminator */
	{ NULL, 0, NULL, 0 }
      };

      int option_index = 0;
      int ch;

      ch = getopt_long(argc, argv, "hi:lv", long_options, &option_index);
      if (ch == -1)
	break;

      switch (ch)
	{
	case 'h':
	  show_help = 1;
	  break;

	case 'i':
	  ipod_path = optarg;
	  break;

	case 'l':
	  show_license = 1;
	  break;

	case 'v':
	  show_version = 1;
	  break;
	}
    }

  if (optind < argc)
    {
      /* Build a copy of argv without ipodfs-specific options. */
      fuse_argc = argc - optind + 1;
      fuse_argv = calloc(fuse_argc, sizeof(char *));
      if (fuse_argv)
	{
	  int i;

	  for (i = 0; i < fuse_argc; ++i)
	    {
	      if (!i)
		fuse_argv[i] = strdup(argv[0]);
	      else
		fuse_argv[i] = strdup(argv[optind + i - 1]);

	      if (!fuse_argv[i])
		{
		  perror(program_name);
		  ok = 0;
		  break;
		}
	    }
	}
      else
	{
	  fuse_argc = 0;
	  ok = 0;
	}
    }

  if (show_help + show_license + show_version > 1)
    {
      fprintf(stderr,
	      "ERROR: Only one of --help, --license and --version can be used\n");
      ok = 0;
    }

  if (!fuse_argc || show_help)
    {
      if (show_help || (!show_help && !show_license && !show_version))
	help(program_name);
      ok = 0;
    }

  if (show_license && !show_help && !show_version)
    {
      license();
      ok = 0;
    }

  if (show_version && !show_help && !show_license)
    {
      version();
      ok = 0;
    }

  if (ok && !ipod_path)
    {
      ipod_path = DEFAULT_IPOD_PATH;
      fprintf(stderr,
	      "WARNING: No iPod path given; using default: %s\n",
	      ipod_path);
    }

  /* Store information about iPod's filesystem for use in ipodfs. */
  if (ok && (statvfs(ipod_path, &ipod_statvfs) == -1))
    {
      perror(program_name);
      ok = 0;
    }

  /* Load up the iTunes DB */
  if (ok)
    itunes = itdb_parse(ipod_path, &err);
  if (err != NULL)
    {
      fprintf (stderr, "Unable to open iTunes database: %s\n", err->message);
      g_error_free (err);
      ok = 0;
    }

  /* Set up initial ipodfs_context */
  if (ok)
    {
      initial_ctx.itunes = itunes;
      initial_ctx.ipod_path = ipod_path;
      memcpy(&initial_ctx.ipod_statvfs, &ipod_statvfs, sizeof(struct statvfs));
    }

  /* Start up FUSE */
  umask(0);
  if (ok && !fuse_main(fuse_argc, fuse_argv, &ipodfs_operations))
    ok = 0;

  /* Clean up */
  if (fuse_argv)
    {
      int i;

      for (i = 0; i < fuse_argc; ++i)
	{
	  if (fuse_argv[i])
	    free(fuse_argv[i]);
	}
    }

  return ok ? EXIT_SUCCESS : EXIT_FAILURE;
}
